from .training import train, regLoss
#from .training import multitaskMSE, multitaskTrain, multitaskLoss
from .training import test
from .training import LangevinOpt
from .networks import FCNet
from .networks import ConvNet


from .classic_tasks import cifar_dataset
from .classic_tasks import oneHotEncoding
from .classic_tasks import mnist_dataset
#from .extra_tasks import emnistABFL_CHIS
#from .extra_tasks import emnistABEL_CHJS
from .synthetic_tasks import synthetic_1hl_dataset



import torch, torchvision, torchvision.transforms as t 
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
from ..NN import *

class random_dataset: 
    def __init__(self, N):
        self.N = N

    def make_data(self, P, Ptest):
        inputs = torch.randn((P, self.N))
        targets = torch.randn(P)
        test_inputs = torch.randn((Ptest, self.N))
        test_targets = torch.randn(Ptest)
        return inputs, targets, test_inputs, test_targets
    
class hmm_dataset:
    def __init__(self, N, D, F, Wt, whichTask, q, eta, rho, actx):
        self.N = N
        self.D = D
        self.side_size = int(np.sqrt(self.N))
        self.save_data = bool("NaN")
        self.randmat_file = f"randmat_N0_{self.N}.txt"
        self.whichTask = whichTask
        self.q = q
        self.eta = eta
        self.rho = rho
        self.actx = actx
        self.Wt = Wt
        self.F = F

    def perturb_teachers(self):
        Wt_perturbed = np.copy(self.Wt)
        n_latent = len(Wt_perturbed)
        #print("latent dimension is", n_latent)
        Wt_perturbed = self.q * Wt_perturbed + np.sqrt(1 - self.q**2) * np.random.normal(0., 1., size = (n_latent))
        return Wt_perturbed
    
    def perturb_features(self):
        F_perturbed = np.copy(self.F)
        n_latent, n = np.shape(F_perturbed)
        n_rows = int(self.rho*n_latent)
        F_perturbed = self.eta * F_perturbed + np.sqrt(1 - self.eta**2) * np.random.normal(0, 1., size = (n_latent, n))
        F_perturbed[:n_rows] = np.random.normal(0, 1., size = (n_rows, n))
        return F_perturbed

    def get_hmm(self, P, P_test):
        if self.whichTask == 'target':
            Wt = self.perturb_teachers()
            F = self.perturb_features()
        else: 
            Wt = self.Wt
            F = self.F
        c_train = np.random.normal(0., 1., size = (P, self.D))
        c_test = np.random.normal(0., 1., size = (P_test, self.D))
        if self.actx == 'tanh':
            x_train = np.tanh(c_train@F/np.sqrt(self.D))
            x_test = np.tanh(c_test@F/np.sqrt(self.D))
        if self.actx == 'relu':
            x_train = relu(c_train@F/np.sqrt(self.D))
            x_test = relu(c_test@F/np.sqrt(self.D))
        y_train = np.sign(c_train @ Wt)
        y_test = np.sign(c_test @ Wt)
        x_train, y_train, x_test, y_test = torch.tensor(x_train, dtype=torch.float), torch.tensor(y_train, dtype=torch.float), torch.tensor(x_test, dtype=torch.float), torch.tensor(y_test, dtype=torch.float)
        return x_train, y_train, x_test, y_test

    def make_data(self, P, P_test, batch_size):

        if self.whichTask not in ["source", "target"]:
            raise Exception("Sorry, no task found!")
        else:
            data, labels, test_data, test_labels = self.get_hmm(P, P_test)
        #print("the dimension of data is ", len(data[0]))
        return data, labels, test_data, test_labels


class synthetic_1hl_dataset: 
    def __init__(self, N, hidden_dim, act, dataSeed = 1234):
        self.N = N
        self.hidden_dim = hidden_dim
        model = FCNet(N, hidden_dim, L=1)
        self.model = model.Sequential(bias = False, act_func=act)
        # Initialize the model parameters
        self.dataSeed = dataSeed
        self.initialize_model()

    def initialize_model(self):
        # Set the model to evaluation mode and initialize with random weights
        self.model.eval()
        with torch.no_grad():
            for param in self.model.parameters():
                nn.init.normal_(param, mean=0, std=1)
    
    def make_data(self, P, Ptest):
        rng = np.random.RandomState(self.dataSeed) 
        inputs = torch.tensor(rng.randn(P, self.N), dtype=torch.float)
        test_inputs = torch.tensor(rng.randn(Ptest, self.N), dtype=torch.float)
        with torch.no_grad():
            targets = self.model(inputs).squeeze()
            test_targets = self.model(test_inputs).squeeze()
        
        return inputs, targets.unsqueeze(1), test_inputs, test_targets.unsqueeze(1)
    



import torch, torchvision, torchvision.transforms as t 
import numpy as np
import torch.nn as nn
import torch.nn.functional as F

def relu(x):
    return x if x > 0 else 0

def getTransforms(self):
        T = t.Compose([
            t.Resize(size = self.side_size), 
            t.ToTensor(), 
            t.Grayscale(), 
            #t.Normalize((0.5),(0.24)),
            t.Lambda(lambda x: torch.flatten(x))
        ])
        return T

def normalizeDataset(data, testData):
    mean = data.mean()
    std = data.std()
    data = data - mean
    testData = testData - mean
    data = data / std
    testData = testData /std
    return data, testData

def accumulateClasses(X, Y, classesTask1, classesTask2):
    X1 = X[Y == classesTask1[0]]
    X2 = X[Y == classesTask2[0]]
    for c in range(1,len(classesTask1)):
        X1 = np.concatenate([X1, X[Y == classesTask1[c]]])
        X2 = np.concatenate([X2, X[Y == classesTask2[c]]])
    Y1, Y2  = np.zeros(len(X1)), np.ones(len(X2))
    Xnew, Ynew = np.concatenate([X1, X2]), np.concatenate([Y1, Y2])
    return Xnew, Ynew

def preprocessDataset(Xtrain, Ytrain, Xtest, Ytest, dataSeed, whichLabels, P,Ptest):
    rng = np.random.RandomState(dataSeed)   
    rp = rng.permutation(len(Ytrain))
    rptest = rng.permutation(len(Ytest))
    Xtrain, Ytrain = Xtrain[rp[:P]], Ytrain[rp[:P]]
    Xtest, Ytest =  Xtest[rptest[:Ptest]], Ytest[rptest[:Ptest]]
    if whichLabels[0] == -1 and whichLabels[1] == 1:
        Ytrain = 2 * Ytrain - 1
        Ytest = 2 * Ytest - 1
    return Xtrain, Ytrain, Xtest, Ytest

def randomProjectDataset(Xtrain, Xtest, D, randomProject):
    N0 = len(Xtrain[0])
    rng2 = np.random.RandomState(randomProject) 
    wmatrix = rng2.random(( N0, D))
    Xtrain = relu(np.dot(Xtrain, wmatrix)/np.sqrt(N0))
    Xtest = relu(np.dot(Xtest, wmatrix)/np.sqrt(N0))
    return Xtrain, Xtest

class emnistABEL_CHJS: 
    def __init__(self, N, whichLabels, randomProject, dataSeed):
        self.N = N
        self.side_size = int(np.sqrt(self.N))
        self.whichLabels = whichLabels
        if randomProject > 0:
            self.projectSeed = randomProject
            self.randomProject = True
        self.dataSeed = dataSeed
    def get_dataset_emnist(self, Xtrain, Ytrain, Xtest, Ytest):
        Xtrain, Ytrain = accumulateClasses(Xtrain, Ytrain, [1,2,5,12], [3,8,10,19])
        Xtest, Ytest = accumulateClasses(Xtest, Ytest, [1,2,5,12], [3,8,10,19])
        Xtrain, Ytrain, Xtest, Ytest = preprocessDataset(Xtrain, Ytrain, Xtest, Ytest, self.dataSeed, self.whichLabels, self.P, self.Ptest)
        Xtrain, Xtest = normalizeDataset(Xtrain, Xtest)
        if self.projectSeed > 0:
            Xtrain, Xtest = randomProjectDataset(Xtrain, Xtest, self.D, self.projectSeed)
        Xtrain, Ytrain, Xtest, Ytest = torch.tensor(Xtrain, dtype=torch.float), torch.tensor(Ytrain, dtype=torch.float), torch.tensor(Xtest, dtype=torch.float), torch.tensor(Ytest, dtype=torch.float)
        return Xtrain, Ytrain, Xtest, Ytest 
    def make_data(self, P, Ptest, batch_size, D):
        self.P, self.Ptest, self.D = P, Ptest, D
        transformDataset = getTransforms(self)
        trainset = torchvision.datasets.EMNIST(root = './data', download = True, split = "letters", train = True,  transform = transformDataset)
        trainloader = torch.utils.data.DataLoader(trainset, batch_size = batch_size, num_workers = 0)
        testset = torchvision.datasets.EMNIST(root = './data', download = True, split = "letters", train = False, transform = transformDataset)
        testloader = torch.utils.data.DataLoader(testset, batch_size = batch_size, num_workers = 0)
        data, labels = next(iter(trainloader))
        Xtest, Ytest = next(iter(testloader))
        data, labels, Xtest, Ytest = self.get_dataset_emnist(data, labels, Xtest, Ytest)
        return data, labels.unsqueeze(1), Xtest, Ytest.unsqueeze(1)
    
class emnistABFL_CHIS: 
    def __init__(self, N, whichLabels, randomProject, dataSeed):
        self.N = N
        self.side_size = int(np.sqrt(self.N))
        self.whichLabels = whichLabels
        if randomProject > 0:
            self.projectSeed = randomProject
            self.randomProject = True
        self.dataSeed = dataSeed
    def get_dataset_emnist(self, Xtrain, Ytrain, Xtest, Ytest):
        Xtrain, Ytrain = accumulateClasses(Xtrain, Ytrain, [1,2,6,12], [3,8,9,19])
        Xtest, Ytest = accumulateClasses(Xtest, Ytest, [1,2,6,12], [3,8,9,19])
        Xtrain, Ytrain, Xtest, Ytest = preprocessDataset(Xtrain, Ytrain, Xtest, Ytest, self.dataSeed, self.whichLabels, self.P, self.Ptest)
        Xtrain, Xtest = normalizeDataset(Xtrain, Xtest)
        if self.projectSeed > 0:
            Xtrain, Xtest = randomProjectDataset(Xtrain, Xtest, self.D, self.projectSeed)
        Xtrain, Ytrain, Xtest, Ytest = torch.tensor(Xtrain, dtype=torch.float), torch.tensor(Ytrain, dtype=torch.float), torch.tensor(Xtest, dtype=torch.float), torch.tensor(Ytest, dtype=torch.float)
        return Xtrain, Ytrain, Xtest, Ytest 
    def make_data(self, P, Ptest, batch_size, D):
        self.P, self.Ptest, self.D = P, Ptest, D
        transformDataset = getTransforms(self)
        trainset = torchvision.datasets.EMNIST(root = './data', download = True, split = "letters", train = True,  transform = transformDataset)
        trainloader = torch.utils.data.DataLoader(trainset, batch_size = batch_size, num_workers = 0)
        testset = torchvision.datasets.EMNIST(root = './data', download = True, split = "letters", train = False, transform = transformDataset)
        testloader = torch.utils.data.DataLoader(testset, batch_size = batch_size, num_workers = 0)
        data, labels = next(iter(trainloader))
        Xtest, Ytest = next(iter(testloader))
        data, labels, Xtest, Ytest = self.get_dataset_emnist(data, labels, Xtest, Ytest)
        return data, labels.unsqueeze(1), Xtest, Ytest.unsqueeze(1)


import numpy as np
from scipy.optimize import fsolve
from .. import kernels
import torch

class CONV_1HL():
    def __init__(self, 
                 Nc     : int,
                 T      : float, 
                 mask   : int,
                 stride : int, 
                 l0     : float = 1., 
                 l1     : float = 1., 
                 act    : str = "erf"
                 ):
        self.Nc, self.l0, self.l1, self.T, self.mask, self.stride = Nc, l0, l1, T, mask, stride
        self.pool = False
        self.kernel, self.kernelTorch = eval(f"kernels.kernel_{act}"), eval(f"kernels.kernel_{act}_torch")
        self.corrNorm = 1 / (self.mask**2 * self.l0)
        self.corrNormPool = 1 / (self.mask**4 * self.l0)
        
    def effectiveAction(self, Q):
        Q = Q.reshape(self.numPatches, self.numPatches)
        rKTemp = torch.zeros((self.P,self.P), dtype = torch.float64)
        for i in range(self.numPatches):
            for j in range(self.numPatches):
                Kij    = torch.tensor(self.CXX[i][j])
                Kii    = torch.tensor(self.CXX[i][i])
                Kjj    = torch.tensor(self.CXX[j][j])
                rKij   = Q[i][j] * kernels.computeKmatrixMultipleCTorch(Kii, Kij, Kjj, self.kernelTorch) / (self.l1 * self.numPatches)
                rKTemp = torch.add(rKTemp , rKij)
        A = rKTemp +  self.T * torch.eye(self.P) 
        invA = torch.inverse(A)
        return (torch.trace(Q) - torch.logdet(Q)
                + (1/self.Nc) * torch.matmul(self.y, torch.matmul(invA, self.y)) 
                + (1/self.Nc) * torch.logdet(A)
                 )
    
    def computeActionGrad(self, Q):
        Q = torch.tensor(Q, dtype = torch.float64, requires_grad = True)
        f = self.effectiveAction(Q)
        f.backward()
        return Q.grad.data.detach().numpy()

    def minimizeAction(self, Q0, minimizationTol = 1e-6, nStep = 100000, gradTol = 1e-3):
        #if isinstance(Q0, float):
        #    Q0 = np.eye(self.numPatches)
        self.optQ = fsolve(self.computeActionGrad, Q0.ravel(), xtol = minimizationTol, maxfev = int(nStep*(self.numPatches**2)))
        self.optQ = np.reshape(self.optQ, (self.numPatches,self.numPatches))
        self.grad = self.computeActionGrad(self.optQ) # gradient of the action 
        self.isClose = np.isclose(self.grad, np.zeros((self.numPatches, self.numPatches)), atol=gradTol) #checks if gradient is smaller than gradTol
        self.converged = self.isClose.all()
        #print("\nis grad close to zero?\n", self.isClose)
    
    #tries optimization nStep times. Each time it starts from the previously-found solution and perturbs it with white noise with variance noiseVar
    def optimize(self, Q0 = 1., nSteps = 5, minimizationTol = 1e-6, nStepSolver = 100000, gradTol = 1e-3, noiseVar = 0.01):
        i = 0
        if isinstance(Q0, float):
            Q0 = np.eye(self.numPatches)
        self.minimizeAction(Q0, minimizationTol = minimizationTol, nStep = nStepSolver, gradTol = gradTol)
        while (not self.converged) and (i < nSteps):
            print(f"solution not found. Try #{i+1}:\n") 
            self.minimizeAction(self.optQ + np.random.randn(*self.optQ.shape)*noiseVar, minimizationTol = minimizationTol, nStep = nStepSolver, gradTol = gradTol)
            i += 1     
    
    def setIW(self):
        self.optQ = np.eye(self.numPatches)
    
    def setPooling(self):
        self.pool = True
        self.optQ = np.ones((self.numPatches, self.numPatches))

    def preprocess(self, X, Y):
        self.P, self.N0 = X.shape
        self.side = int(np.sqrt(self.N0))
        self.X, self.Y = X, Y
        self.numPatches = int(self.N0 / (self.mask**2))
        patchX = np.array([kernels.divide2dImage(X[mu].reshape(self.side,self.side), self.mask) for mu in range(self.P)])
        self.Xtrain = np.zeros(self.numPatches, dtype = object)
        for i in range(self.numPatches):
            self.Xtrain[i] = patchX[:,i].reshape(self.P, -1) #the i-th element of Xtrain is an object with dimension (P, m^2), where m^2 is the number of elements in the single patch
        CXX = np.zeros((self.numPatches,self.numPatches), dtype = object)
        for i in range(self.numPatches):
            for j in range(i, self.numPatches):
                CXX[i][j] = np.dot(self.Xtrain[i], self.Xtrain[j].T) * self.corrNorm
                CXX[j][i] = CXX[i][j].T
        self.CXX = CXX
        self.y = Y.squeeze().to(torch.float64)
        self.y.requires_grad = False

    def computeTestsetKernels(self, Xtest):
        self.Ptest = len(Xtest)
        patchXtest = np.array([kernels.divide2dImage(Xtest[mu].reshape(self.side,self.side), self.mask) for mu in range(self.Ptest)])
        self.Xtest = np.zeros((self.numPatches), dtype = object ) # Each i element of Xtest is a P* times m^2 matrix, where m^2 is the number of elements in one patch 
        for i in range(self.numPatches):
            self.Xtest[i] = patchXtest[:,i].reshape(self.Ptest, -1) 
        C00 = np.zeros((self.numPatches, self.numPatches), dtype = object) # Each (i,j) element of the matrix C00 is a P* x P* matrix
        CX0 = np.zeros((self.numPatches, self.numPatches), dtype = object) # Each (i,j) element of the matrix CX0 is a P x P* matrix
        C0X = np.zeros((self.numPatches, self.numPatches), dtype = object)
        for i in range(self.numPatches):
            for j in range(self.numPatches):
                C00[i][j] = np.dot(self.Xtest[i], self.Xtest[j].T)  * self.corrNorm
                CX0[i][j] = np.dot(self.Xtrain[i], self.Xtest[j].T) * self.corrNorm
                C0X[i][j] = np.dot(self.Xtest[i], self.Xtrain[j].T) * self.corrNorm 
        self.CX0, self.C00, self.C0X = CX0, C00, C0X
    
    def predict(self, Xtest):
        self.computeTestsetKernels(Xtest)
        self.orderParam = self.optQ / (self.l1 * self.numPatches)
        if self.pool:
            self.orderParam /= self.numPatches
        rK0XTemp = np.zeros((self.Ptest, self.P), dtype = np.float64)
        rK0Temp  = np.zeros((self.Ptest, self.Ptest), dtype = np.float64) #to check why this is a matrix and then becomes vector
        rKTemp   = np.zeros((self.P, self.P), dtype = np.float64)
        KijTemp, K0XijTemp, K0ijTemp = [], [], []
        for i in range(self.numPatches):
            for j in range(self.numPatches):
                KXii  = self.CXX[i,i].diagonal() #diagonal in train data (mu, mu)
                KXXij = self.CXX[i,j]
                KXjj  = self.CXX[j,j].diagonal() #diagonal in train data (mu, mu)
                K0Xij = self.C0X[i,j]
                K0ii  = self.C00[i,i].diagonal() #diagonal in test data (mu*, mu*)
                K0ij  = self.C00[i,j].diagonal() #diagonal in test data (mu*, mu*)
                K0jj  = self.C00[j,j].diagonal() #diagonal in test data (mu*, mu*)
                Ktrain = self.kernel(KXii[:, None], KXXij, KXjj[None, :])
                Kmix   = self.kernel(K0ii[:, None], K0Xij, KXjj[None, :])
                Ktest  = self.kernel(K0ii, K0ij, K0jj)
                KijTemp.append(Ktrain)
                K0ijTemp.append(Ktest)
                K0XijTemp.append(Kmix)
                rKij   = self.orderParam[i,j] * Ktrain
                rK0Xij = self.orderParam[i,j] * Kmix
                rK0ij  = self.orderParam[i,j] * Ktest
                rKTemp   = np.add(rKTemp, rKij)
                rK0Temp  = np.add(rK0Temp, rK0ij)
                rK0XTemp  = np.add(rK0XTemp, rK0Xij)
        self.rK, self.rK0, self.rK0X = rKTemp, rK0Temp, rK0XTemp
        self.Kij, self.K0Xij, self.K0ij = KijTemp, K0XijTemp, K0ijTemp
        A = self.rK + self.T * np.eye(self.P)
        invK = np.linalg.inv(A)
        self.K0_invK = np.matmul(self.rK0X, invK)
        self.Ypred = np.matmul(self.K0_invK, self.Y)
        return self.Ypred

    def averageLoss(self, Ytest):
        bias = Ytest - self.Ypred
        var = self.rK0 - np.matmul(self.K0_invK, self.rK0X.T).diagonal()
        predLoss = bias**2 + var
        return predLoss.mean().item(), (bias**2).mean().item(), var.mean().item()
    
class validation_CONV_1HL():
    def __init__(self, Kij, K0Xij, K0ij, Yval, Ytrain,  T, l0 = 1., l1 = 1., act = "erf", epsilon = 0.01):
        self.Kij, self.K0ij, self.K0Xij = torch.tensor(Kij, requires_grad = False), torch.tensor(K0ij, requires_grad = False), torch.tensor(K0Xij, requires_grad = False)
        self.numPatches = int(np.sqrt(len(Kij)))
        self.P = len(self.Kij[0])
        self.Pval = len(self.K0ij[0])
        self.l0, self.l1, self.T, self.act = l0, l1, T, act
        self.pool = False
        self.Yval = Yval.to(torch.float64)
        self.Ytrain = Ytrain.to(torch.float64)
        self.kernel, self.kernelTorch = eval(f"kernels.kernel_{act}"), eval(f"kernels.kernel_{act}_torch")
        self.epsilon = epsilon #learning rate
        
    def bias(self, Q):
        rK0XTemp = torch.zeros((self.Pval, self.P), dtype = torch.float64)
        rK0Temp  = torch.zeros((self.Pval,self.Pval), dtype = torch.float64)
        rKTemp   = torch.zeros((self.P, self.P), dtype = torch.float64)
        for i in range(self.numPatches):
            for j in range(self.numPatches):
                index = int(self.indexMat[i,j])             
                rKij   = Q[index] * self.Kij[i*self.numPatches+j] / (self.l1 * self.numPatches)
                rK0Xij = Q[index] * self.K0Xij[i*self.numPatches+j] / (self.l1 * self.numPatches)
                rK0ij  = Q[index] * self.K0ij[i*self.numPatches+j] / (self.l1 * self.numPatches)
                rKTemp   = torch.add(rKTemp, rKij)
                rK0Temp  = torch.add(rK0Temp, rK0ij)
                rK0XTemp  = torch.add(rK0XTemp, rK0Xij)
        self.rK, self.rK0, self.rK0X = rKTemp, rK0Temp, rK0XTemp 
        A = rKTemp +  self.T * torch.eye(self.P) 
        invA = torch.inverse(A)
        self.K0_invK = torch.matmul(self.rK0X, invA)
        self.Ypred = torch.matmul(self.K0_invK, self.Ytrain)
        return torch.mean((self.Yval - self.Ypred)**2)
    
    def computeBiasGrad(self, Q):
        Q = torch.tensor(Q, dtype = torch.float64, requires_grad = True)
        f = self.bias(Q)
        f.backward()
        return Q.grad.data.detach().numpy()

    def minimizeBias(self, Q0 = 1., gradTol = 1e-3):
        if isinstance(Q0, float):
            Q0 = np.eye(self.numPatches)
        numVariables = int(self.numPatches*(self.numPatches+1)/2)
        self.indexMat = np.zeros((self.numPatches, self.numPatches))
        Qstart = np.zeros(numVariables)
        count = 0
        for i in range(self.numPatches): 
            for j in range(i, self.numPatches): 
                self.indexMat[i,j] = count
                self.indexMat[j,i] = count
                Qstart[count] = Q0[i,j]
                count +=1
        optQ = Qstart - self.epsilon * self.computeBiasGrad(Qstart)
        self.optQ = np.zeros((self.numPatches, self.numPatches))
        count=0
        for i in range(self.numPatches): 
            for j in range(i, self.numPatches): 
                self.optQ[i,j] = optQ[count]
                self.optQ[j,i] = optQ[count]
                count += 1
        self.grad = self.computeBiasGrad(optQ) # gradient of the action 
        self.isClose = np.isclose(self.grad, np.zeros(numVariables), atol=gradTol) #checks if gradient is smaller than gradTol
        self.converged = self.isClose.all()
        #print("\nis grad close to zero?\n", self.isClose)



import os, os.path,time,argparse
import numpy as np
import torch, torch.nn as nn
import torch.backends.cudnn as cudnn
import torch.nn.init as init
import torch.optim as optim
from torch import nn
from torch.optim import Optimizer
import torch.nn.functional as F

class Erf(torch.nn.Module):
    def __init__(self):
        super().__init__()
    def forward(self, x):
        return torch.erf(x)

class Id(torch.nn.Module):
    def __init__(self):
        super().__init__()
    def forward(self, x):
        return x

class Norm(torch.nn.Module):
    def __init__(self, norm):
        super().__init__()
        self.norm = norm
    def forward(self, x):
        return x/self.norm
    
def relu(x):
    return x if x > 0 else 0

class FCNet: 
    def __init__(self, 
                 N0 : int, #input dimension
                 N1 : int, #number of hidden layer units
                 L : int, #number of hidden layers. depth = L+1
                 bias : bool = False, #if True, add bias for each layer
                 act : str = "erf" #activation function
                 ):
        self.N0, self.N1, self.L, self.act, self.bias = N0, N1, L, act, bias 

    def Sequential(self):
        if self.act == "relu":
            act = nn.ReLU()
        elif self.act == "erf":
            act = Erf()
        elif self.act == "id":
            act = Id()
        modules = []
        first_layer = nn.Linear(self.N0, self.N1, bias=self.bias)
        init.normal_(first_layer.weight, std = 1)
        if self.bias:
            init.constant_(first_layer.bias,0)
        modules.append(Norm(np.sqrt(self.N0)))
        modules.append(first_layer)
        for l in range(self.L-1): 
            modules.append(act)
            modules.append(Norm(np.sqrt(self.N1)))
            layer = nn.Linear(self.N1, self.N1, bias = self.bias)
            init.normal_(layer.weight, std = 1)
            if self.bias:
                init.normal_(layer.bias,std = 1)
            modules.append(layer)
        modules.append(act)
        modules.append(Norm(np.sqrt(self.N1)))
        last_layer = nn.Linear(self.N1, 1, bias=self.bias)  
        init.normal_(last_layer.weight, std = 1) 
        if self.bias:
                init.normal_(last_layer.bias,std = 1)
        modules.append(last_layer)
        sequential = nn.Sequential(*modules)
        print(f'\nThe network has {self.L} dense hidden layer(s) of size {self.N1} with {self.act} actviation function', sequential)
        return sequential
    
class ConvNet:
    def __init__(self, N0 : int,  #input size, necessary to compute normalization
                 Nc : int,  # number of channels in internal layers
                 mask : int, #mask
                 stride : int, #striding of convolution
                 bias : bool = False, #if True, add bias for each layer
                 act :str = "erf", #activation function
                 inputChannels : int = 1 #number of inputs channel, defaults for black and white image
                 ):
        self.Nc, self.N0, self.inputChannels, self.mask, self.stride, self.bias, self.act = Nc, N0, inputChannels, mask, stride, bias, act

    def Sequential(self):
        if self.act == "relu":
            act = nn.ReLU()
        elif self.act == "erf":
            act = Erf()
        modules = []
        # First convolutional layer
        first_layer = nn.Conv2d(self.inputChannels, self.Nc, self.mask, stride = self.stride, bias=self.bias)
        init.normal_(first_layer.weight, std=1)
        if self.bias:
            init.constant_(first_layer.bias, 0)
        modules.extend([Norm(self.mask), first_layer, act])
        # Flatten the tensor before the fully connected layer
        modules.append(nn.Flatten())
        with torch.no_grad(): 
            dummy_net_sequential = nn.Sequential(*modules)
            out = dummy_net_sequential(torch.randn(self.inputChannels,int(np.sqrt(self.N0)),int(np.sqrt(self.N0))).unsqueeze(0))
            FC_params = len(out.flatten())  # number of parameters in the last layer   
        # Fully connected layer
        last_layer = nn.Linear(FC_params, 1, bias=self.bias)
        init.normal_(last_layer.weight, std=1)
        if self.bias:
            init.constant_(last_layer.bias, 0)
        modules.append(Norm(np.sqrt(FC_params)))
        modules.append(last_layer)
        sequential = nn.Sequential(*modules)
        print(f'\nThe network has 1 convolutional hidden layer with {self.Nc} kernels of size {self.mask} and {self.act} activation function', sequential)
        return sequential


import os, os.path, time, argparse
import numpy as np
import torch, torch.nn as nn
import torch.backends.cudnn as cudnn
import torch.nn.init as init
import torch.optim as optim
from torch import nn
from torch.optim import Optimizer


class LangevinOpt(optim.Optimizer):
    def __init__(self, model: nn.Module, lr, temperature):
        defaults = {
            'lr': lr, 
            'temperature': temperature
        }
        groups = []
        for i, layer in enumerate(model.children()):
            groups.append({'params': layer.parameters(), 
                           'lr': lr, 
                           'temperature': temperature})
        super(LangevinOpt, self).__init__(groups, defaults)
        
    @torch.no_grad()
    def step(self, closure=None):
        for group in self.param_groups:
            learning_rate = group['lr']
            temperature = group['temperature']
            for parameter in group['params']:
                if parameter.grad is None:
                    continue
                d_p = torch.randn_like(parameter) * (2*learning_rate*temperature)**0.5
                d_p.add_(parameter.grad, alpha=-learning_rate)
                parameter.add_(d_p)

def multitaskTrain(Snet, Tnet, source_data, target_data, source_labels, target_labels, criterion, Soptimizer, Toptimizer, args):
    Snet.train()
    Tnet.train()
    Soptimizer.zero_grad()
    Toptimizer.zero_grad()
    source_outputs = Snet(source_data)
    target_outputs = Tnet(target_data)
    loss = criterion(source_outputs.float(), target_outputs.float(), source_labels.float(), target_labels.float(), Snet, Tnet, args)
    loss.backward()
    Soptimizer.step()
    Toptimizer.step()
    return loss.item() 

def test(net, test_data, test_labels, criterion):
        net.eval()
        with torch.no_grad():
                outputs = net(test_data)
                loss = criterion(outputs, test_labels) 
        return loss.item()

def multitaskMSE(Snet, Tnet, Sinputs, Tinputs, Stargets, Ttargets, criterion, args):
        Snet.eval()
        Tnet.eval()
        with torch.no_grad():
                Soutputs = Snet(Sinputs)
                Toutputs = Tnet(Tinputs)
                Sloss = criterion(Soutputs, Stargets) 
                Tloss = criterion(Toutputs, Ttargets) 
        loss = Sloss + Tloss 
        return loss.item()/(args.Ps + args.Pt)

def multitaskLoss(source_output, target_output, source_labels, target_labels, Snet, Tnet, args):
    loss = 0
    for i in range(len(Snet)):
        if i == len(Snet)-1:
            loss += (0.5*args.lambda1*args.T)*(torch.linalg.matrix_norm(Snet[i].weight)**2)
            loss += (0.5*args.lambda1*args.T)*(torch.linalg.matrix_norm(Tnet[i].weight)**2)
        else:
            if isinstance(Snet[i], nn.Linear):
                #print("hello")
                #loss += (0.5*(args.lambda0-args.gamma)*args.T)*(torch.linalg.matrix_norm(Snet[i].weight)**2)
                #loss += (0.5*(args.lambda0-args.gamma)*args.T)*(torch.linalg.matrix_norm(Tnet[i].weight)**2)
                loss += (0.5*(args.lambda0)*args.T)*(torch.linalg.matrix_norm(Snet[i].weight)**2)
                loss += (0.5*(args.lambda0)*args.T)*(torch.linalg.matrix_norm(Tnet[i].weight)**2)
                #loss += (0.5*args.gamma*args.T)*(torch.linalg.matrix_norm(Snet[i].weight-Tnet[i].weight)**2)
                loss -= (0.5*args.gamma*args.T)*(torch.dot(Snet[i].weight.flatten(),Tnet[i].weight.flatten()))
    loss += 0.5*torch.sum((source_output - source_labels)**2)
    loss += 0.5*torch.sum((target_output - target_labels)**2)
    return loss

def computeNormsAndOverlaps(Snet,Tnet):
    i = len(Snet)-1
    Snorm = torch.linalg.matrix_norm(Snet[i].weight)
    Tnorm = torch.linalg.matrix_norm(Tnet[i].weight)
    overlap = torch.dot(Snet[i].weight.flatten(),Tnet[i].weight.flatten())
    return Snorm, Tnorm, overlap

def train(net,data, labels, criterion, optimizer,T,lambda0,lambda1):
    net.train()
    inputs, targets = data, labels#).unsqueeze(1)
    optimizer.zero_grad()
    outputs = net(inputs)
    #loss = criterion(outputs.float(),targets.float())
    loss = criterion(outputs.float(), targets.float(),net,T,lambda0,lambda1)
    loss.backward()
    optimizer.step()
    return loss.item() 

def regLoss(output, target,net,T,lambda0,lambda1):
    loss = 0
    for i in range(len(net)):
        if i == len(net)-1:
            loss += (0.5*lambda1*T)*(torch.linalg.matrix_norm(net[i].weight)**2)
        else:
            if (isinstance(net[i],nn.Linear) or isinstance(net[i],nn.Conv2d)):
                loss += (0.5*lambda0*T)*(torch.linalg.norm(net[i].weight.flatten())**2)
    loss += 0.5*torch.sum((output - target)**2)
    return loss


import numpy as np
from typing import Union, List
from scipy.optimize import minimize
from . import rkgp

def RKGP(L: int, 
         width: int, 
         D: int, 
         T: float, 
         priors: List[float] = [], 
         act: str = "erf", 
         model_type: str = "FC", 
         mask : int = 1,
         stride : int = 1,
         kernel_type: str = "vanilla"):
    """
    Constructs a Gaussian Process model.

    Parameters:
    - L (int): Number of hidden layers. Must be an integer.
    - width (int): Size of the hidden layer (number of channels if model is convolutional). Must be an integer.
    - D (int): Number of network outputs. Must be an integer greater than 0.
    - T (float): Temperature parameter.
    - priors (list of float, optional): Gaussian priors for each layer. Defaults to None.
    - act (str, optional): Activation function. Defaults to "erf".
    - model_type (str, optional): Type of model. Defaults to "FC".
    - kernel_type (str, optional): Kernel type. Defaults to "best". Setting kernel_type = "vanilla" does not implement first order corrections to the proportional regime

    Returns:
    - model: The constructed RKGP model that is equivalent to the specified network.
    """
    
    morespecs = ""
    assert (isinstance(L, int) and L > 0 ), "Number of layers must be integer and greater than 0."
    initPriors = np.ones(L)

    assert width > 0 , "Size of hidden layer(s) must be greater than 0."
    argsDict = {"depth of the network (L+1)": L+1, "size of hidden layer(s)": width, "number of outputs": D, "temperature": T, "activation function": act}

    if L == 1: 
        netName = "1HL"
        if len(priors) != 0:
            l0, l1 = priors[0], priors[1]
        else:
            l0, l1 = initPriors # if the user choose no priors
        argsDict["Gaussian priors"] = (l0, l1)
        assert (isinstance(D, int) and D>=1), "Number of outputs must be integer and greater than 0."
        if D == 1: 
            args = [width, T, l0, l1, act]
            argNames = {"N1":width, "T": T, "l0": l0, "l1": l1, "act": act}
        else: 
            morespecs = "_multiclass"
            args = [width, D, T, l0, l1, act]
            argNames = {"N1":width, "D": D, "T": T, "l0": l0, "l1": l1, "act": act}
        if model_type == "CONV": 
            assert D == 1,  "No model for shallow convolutional networks with multiple outputs."
            args = [width, T, mask, stride, l0, l1, act ]
            argNames = {"Nc":width, "T": T, "mask":mask, "stride":stride,"l0": l0, "l1": l1, "act": act}
    elif L > 1: 
        argsDict[ "Gaussian priors"] = priors
        netName = "deep"
        assert D == 1, "No model for deep networks with multiple outputs."
        assert model_type != "CONV", "No model for deep convolutional networks."
        args = [L, width, T, priors, act]
        argNames = {"L": L, "N1":width, "T": T, "priors": priors, "act": act}
    if kernel_type == "best":
        assert netName == "1HL", "No model with corrections to proportional regime for deep network."
        assert D == 1, "No model with corrections to proportional regime for networks with multiple outputs."
        assert model_type != "CONV", "No model with corrections to proportional regim for convolutional networks."
        morespecs = morespecs + "_corrected"
    elif kernel_type != "vanilla": 
        raise ValueError("Invalid kernel type.")
    
    if model_type == "CONV":
        argsDict["type of hidden layers"] = "convolutional"
        argsDict["mask"] = mask
        argsDict["stride"] = stride
    elif model_type == "FC":
        argsDict["type of hidden layers"] = "fully-connected"

    for name, arg in argsDict.items():
        print(f"{name}: {arg}")
    argStr = ", ".join([f"{key}={value}" for key, value in argNames.items()])
    print(f"\nmodel name: rkgp.{model_type}_{netName}{morespecs} \nmodel args: {argStr} \n")
    

    model = eval(f"rkgp.{model_type}_{netName}{morespecs}")
    return model(*args)



import numpy as np
from scipy.optimize import minimize
from scipy.optimize import fsolve
from .. import kernels
import torch

## DEBUGGED VERSION, PASSED ALL TESTS 01/07/2024
class FC_1HL():
    def __init__(self, 
                 N1   : int, 
                 T    : float, 
                 l0   : float = 1.0,
                 l1   : float = 1.0,
                 act  : str = "erf", 
                 bias : bool = False):
        self.N1, self.l0, self.l1, self.T = N1, l0, l1, T
        self.kernel = eval(f"kernels.kernel_{act}")
        if bias: 
            self.kernel = eval(f"kernels.kernel_{act}_bias")

    def effectiveAction(self, x):
        A = self.T * np.identity(self.P) + (x/self.l1) * self.diagK
        invA = np.linalg.inv(A)
        return ( x - np.log(x)
            + (1/self.N1) * np.sum(np.log(self.T + x * self.eigvalK / self.l1))
            + (1/self.N1) * np.dot(self.yT, np.dot(invA, self.yT)) )

    def optimize(self, x0 = 1.): # x0 is initial condition
        optQ = minimize(self.effectiveAction, x0, bounds = ((1e-8,np.inf),) , tol=1e-12)
        self.optQ = (optQ.x).item()
        assert self.optQ > 0 , "Unphysical solution found (Q is negative)."

    def setIW(self):
        self.optQ = 1

    def preprocess(self, Xtrain, Ytrain):
        self.P, self.N0 = Xtrain.shape
        self.corrNorm = 1/(self.N0*self.l0)
        self.C = np.dot(Xtrain, Xtrain.T) * self.corrNorm
        self.Xtrain = Xtrain
        self.Ytrain = Ytrain
        self.CX = self.C.diagonal()
        self.K = kernels.computeKmatrix(self.C, self.kernel)
        self.eigvalK, eigvecK = np.linalg.eig(self.K)
        self.diagK = np.diagflat(self.eigvalK)
        self.Udag = eigvecK.T
        self.yT = np.matmul(self.Udag, Ytrain.squeeze())

    def computeTestsetKernels(self, Xtest):
        self.Ptest = len(Xtest)
        self.C0 = np.dot(Xtest, Xtest.T).diagonal() * self.corrNorm
        self.C0X = np.dot(Xtest, self.Xtrain.T) * self.corrNorm
        self.K0 =  self.kernel(self.C0, self.C0, self.C0) 
        self.K0X = self.kernel(self.C0[:,None], self.C0X, self.CX[None, :])
    
    def predict(self, Xtest):
        self.computeTestsetKernels(Xtest)
        self.orderParam = self.optQ / self.l1
        A = self.orderParam * self.K + (self.T) * np.eye(self.P)
        invK = np.linalg.inv(A)
        self.rK0X = self.orderParam * self.K0X
        self.K0_invK = np.matmul(self.rK0X, invK)
        self.Ypred =  np.dot(self.K0_invK, self.Ytrain)
        return self.Ypred
    
    def averageLoss(self, Ytest):
        self.rK0 = self.orderParam * self.K0 
        bias = Ytest - self.Ypred 
        var = self.rK0 - np.sum(self.K0_invK * self.rK0X, axis=1)
        predLoss = bias**2 + var 
        return predLoss.mean().item(), (bias**2).mean().item(), var.mean().item()

## FROM HERE  
class FC_1HL_corrected():
    def __init__(self, N1, T, l0 = 1.0, l1 = 1.0, act = "erf"):
        self.N1 = N1
        self.l0 = l0
        self.l1 = l1 
        self.T = T
        self.kernel = eval(f"kernels.kernel_{act}")

    def effectiveAction(self, x):
        tildeK = self.T * np.identity(self.P) + (x/self.l1) * self.diagK
        invA = np.linalg.inv(tildeK)
        return ( x - np.log(x)
            + (1/self.N1) * np.sum(np.log(self.T + x * self.eigvalK / self.l1))
            + (1/self.N1) * np.dot(self.yT, np.dot(invA, self.yT)) )

    def optimize(self, x0 = 1.): # x0 is initial condition
        optQ = minimize(self.effectiveAction, x0, bounds = ((1e-8,np.inf),) , tol=1e-12)
        self.optQ = (optQ.x).item()
        assert self.optQ > 0 , "Unphysical solution found (Q is negative)."
    
    def setIW(self):
        self.optQ = 1

    def preprocess(self, Xtrain, Ytrain):
        self.P, self.N0 = Xtrain.shape
        self.corrNorm = 1/(self.N0*self.l0)
        self.C = np.dot(Xtrain, Xtrain.T) * self.corrNorm
        self.Xtrain = Xtrain
        self.Ytrain = Ytrain
        self.CX = self.C.diagonal()
        self.K = kernels.computeKmatrix(self.C, self.kernel)
        self.eigvalK, eigvecK = np.linalg.eig(self.K)
        self.diagK = np.diagflat(self.eigvalK)
        self.Udag = eigvecK.T
        self.yT = np.matmul(self.Udag, Ytrain.squeeze())

    def computeTestsetKernels(self, Xtest):
        self.Ptest = len(Xtest)
        self.C0 = np.dot(Xtest, Xtest.T).diagonal() * self.corrNorm
        self.C0X = np.dot(Xtest, self.Xtrain.T) * self.corrNorm
        self.K0 =  self.kernel(self.C0, self.C0, self.C0) 
        self.K0X = self.kernel(self.C0[:,None], self.C0X, self.CX[None, :])
    
    def predict(self, Xtest):
        self.computeTestsetKernels(Xtest)
        self.orderParam = self.optQ / self.l1
        tildeK = self.orderParam * self.K + (self.T) * np.eye(self.P)
        invK = np.linalg.inv(tildeK)
        self.corrTildeK = self.orderParam * self.K + (self.T) * np.eye(self.P)
        self.tildeY = np.matmul(invK, np.outer(self.Ytrain, self.Ytrain))
        self.corrToK = np.matmul(np.matmul(self.tildeY -np.eye(self.P), invK ), self.K)
        self.corrK = np.matmul(self.K, np.eye(self.P)+ self.orderParam * self.corrToK/self.N1)
        self.corrK0X =  np.matmul(self.K0X, np.eye(self.P)+ self.orderParam * self.corrToK/self.N1)
        self.corrToK0 = np.matmul( invK , self.K0X.T)
        self.corrK0 = self.K0 + (self.orderParam/self.N1) * (-np.matmul(self.K0X, self.corrToK0)+np.matmul(self.K0X, np.matmul(self.tildeY,self.corrToK0)))
        self.corrTildeK = self.orderParam * self.corrK + (self.T) * np.eye(self.P)
        invCorrK = np.linalg.inv(self.corrTildeK)
        self.rK0X = self.orderParam * self.corrK0X
        self.K0_invK = np.matmul(self.rK0X, invCorrK)
        self.Ypred =  np.dot(self.K0_invK, self.Ytrain)
        return self.Ypred
    
    def averageLoss(self, Ytest):
        self.rK0 = self.orderParam * self.corrK0 
        bias = Ytest - self.Ypred 
        var = self.rK0 - np.sum(self.K0_invK * self.rK0X, axis=1)
        predLoss = bias**2 + var 
        return predLoss.mean().item(), (bias**2).mean().item(), var.mean().item()
    

class FC_1HL_nonodd(): # the order parameter associated with average norm is called x, the other one si called Bx
    def __init__(self, N1, T, l0 = 1.0, l1 = 1.0, act = "relu"):
        self.N1 = N1
        self.l0 = l0
        self.l1 = l1 
        self.T = T
        self.kernel = eval(f"kernels.kernel_{act}")
        self.mean = eval(f"kernels.mean_{act}")

    def effectiveAction(self, Q):
        x = Q[0]
        Bx = Q[1]
        rK = self.K * x/self.l1
        rKm = (x - (1/(1+Bx)))/self.l1 
        A = rK - rKm +  self.T * torch.eye(self.P) 
        invA = torch.inverse(A)
        return ( - x*Bx 
                + torch.log(1 + Bx)
                + (1/self.N1) * torch.matmul(self.Ytrain, torch.matmul(invA, self.Ytrain)) 
                + (1/self.N1) * torch.logdet(A)
                 )

    def computeGrad(self, x):
        xTemp = torch.tensor(x, dtype = torch.float64, requires_grad = True)
        f = self.effectiveAction(xTemp)
        f.backward(retain_graph=True)
        return xTemp.grad.detach().numpy()   
    
    def preprocess(self, Xtrain, Ytrain):
        self.Xtrain = Xtrain
        self.P , self.N0 = Xtrain.shape
        self.corrNorm = 1/(self.l0* self.N0)
        self.Ytrain = torch.tensor(Ytrain.squeeze()).type('torch.DoubleTensor')
        #self.Ytrain.requires_grad = False
        self.C = np.dot(Xtrain,Xtrain.T) * self.corrNorm
        self.CX = self.C.diagonal()
        self.K = torch.tensor(kernels.computeKmatrix(self.C, self.kernel), dtype = torch.float64, requires_grad = False)
        self.Km = torch.tensor(self.mean(self.CX), dtype = torch.float64, requires_grad = False)
        #Km =  torch.tensor(kernel(m), dtype = torch.float64, requires_grad = False) 

    def optimize(self, x0 = [1,0.3]): # the order parameter associated with average norm is called x, the other one si called Bx
        assert len(x0) == 2, "I need initial conditions for two variables. Choose a list with lenght = 2."
        self.optQ = fsolve(self.computeGrad, x0, xtol = 1e-8)
        self.Q = self.optQ[0]
        self.bQ = self.optQ[1]
        assert self.Q > 0, "Unphysical solution found (Q is negative)."
        #print(f"\nPoint where the gradient is approximately zero: {self.optQ}")
        isClose = np.isclose(self.computeGrad(self.optQ), [0.0, 0.0]) 
        self.converged = isClose.all()
        assert self.converged, "Wrong solution found, gradient is not 0. Try changing initial condition."
        #print("\nis exact solution close to zero?", isClose) 
    
    def setIW(self):
        self.Q = 1.
        self.bQ = 0.
    
    def computeTestsetKernels(self, Xtest):
        self.Ptest = len(Xtest)
        self.C0 = np.dot(Xtest, Xtest.T).diagonal() * self.corrNorm
        self.C0X = np.dot(Xtest, self.Xtrain.T) * self.corrNorm
        self.K0 =  self.kernel(self.C0, self.C0, self.C0)
        self.Km0 = self.mean(self.C0)
        self.K0X = self.kernel(self.C0[:,None], self.C0X, self.CX[None, :])
    
    def predict(self, Xtest):
        self.computeTestsetKernels(Xtest)
        rK = self.K * self.Q/self.l1
        rKm = (self.Q - (1/(1+self.bQ))) * np.outer(self.Km, self.Km)/self.l1
        A = rK - rKm +  self.T * torch.eye(self.P) 
        invK = np.linalg.inv(A)
        rK0X = self.K0X * self.Q /self.l1
        rK0Xm = (self.Q - (1/(1+self.bQ))) * np.outer(self.Km, self.Km0) /self.l1
        self.rK0X = rK0X - rK0Xm
        self.K0_invK = np.matmul(self.rK0X, invK)
        self.Ypred =  np.dot(self.K0_invK, self.Ytrain)
        return self.Ypred
    
    def averageLoss(self, Ytest):
        rK0 = self.K0 * self.Q/self.l1 
        rK0m = (self.Q - (1/(1+self.bQ))) * np.outer(self.Km0, self.Km0) /self.l1
        self.rK0 = rK0 - rK0m
        bias = Ytest - self.Ypred 
        var = self.rK0 - np.sum(self.K0_invK * self.rK0X, axis=1)
        predLoss = bias**2 + var 
        return predLoss.mean().item(), (bias**2).mean().item(), var 


import numpy as np
from scipy.optimize import fsolve
import torch
from torch.autograd import Variable 
from .. import kernels
from .. import tasks
## DEBUGGED VERSION. PASSED ALL TESTS ON 01/07/2024

class FC_1HL_multiclass():
    def __init__(self, 
                 N1 : int, 
                 D  : int, 
                 T  : float, 
                 l0 : float = 1.0, 
                 l1 : float = 1.0, 
                 act : str = "erf", 
                 oneHot : bool = False
                 ):
        self.N1, self.l0, self.l1, self.T, self.D, self.oneHot = N1, l0, l1, T, D, oneHot
        self.kernelTorch = eval(f"kernels.kernel_{act}_torch")
        self.kernel = eval(f"kernels.kernel_{act}")
        self.numOfVariables = int(self.D*(self.D+1)/2)
        
    def effectiveAction(self, *args):
        Q = torch.zeros((self.D, self.D))
        count = 0;
        for i in range(self.D):
            for j in range(i, self.D):
                Q[i][j] = args[count]
                Q[j][i] = Q[i][j]
                count = count + 1;
        rK = torch.zeros((self.D*self.P, self.D*self.P))
        for i in range(self.D):
            for j in range(self.D):
                rK[i*self.P:i*self.P+self.P,j*self.P:j*self.P+self.P] = Q[i,j]*self.D/self.l1        
        A = self.T * torch.eye(self.D * self.P) + rK
        return torch.trace(Q) - torch.logdet(Q) + (1/self.N1) * torch.logdet(A) + (1/self.N1)*torch.matmul(self.y,torch.matmul(torch.inverse(A),self.y))
    
    def computeActionGrad(self, *args):
        variables = [Variable(torch.tensor(arg, requires_grad=True, dtype=torch.float), requires_grad=True) for arg in args]
        f = self.effectiveAction(*variables)
        f.backward()
        return [var.grad.data.detach().numpy() if var.grad is not None else 0 for var in variables]

    def gradientEq(self, vars):
        grads = self.computeActionGrad(*vars)
        return grads

    def fromQtoVariables(self, matrix):
        variables = np.zeros(self.numOfVariables)
        count = 0;
        for i in range(self.D):
            for j in range(i, self.D):
                variables[count] = matrix[i,j]
                count = count + 1;
        return variables

    def fromVariablesToQ(self, var):
        matrix = np.zeros((self.D, self.D))
        count = 0;
        for i in range(self.D):
            for j in range(i, self.D):
                matrix[i][j] = var[count]
                matrix[j][i] = matrix[i][j]
                count = count + 1;
        return matrix

    def optimize(self, x0 = 1):
        if isinstance(x0,(float,int)):
            x0 = np.eye(self.D)
        self.optVar = fsolve(self.gradientEq, self.fromQtoVariables(x0))
        self.optQ = self.fromVariablesToQ(self.optVar)
        print(self.gradientEq(self.optVar))
        isClose = np.isclose(self.gradientEq(self.optVar), np.zeros(self.numOfVariables), atol=1e-05) 
        self.converged = isClose.all()
        #print("\nis exact solution close to zero?", isClose)   
        #print(f"optQ is {self.optQ}")
    
    def setIW(self):
        self.optQ = np.eye(self.D)

    def preprocess(self, Xtrain, Ytrain):
        self.P, self.N0 = Xtrain.shape
        self.corrNorm = 1/(self.N0 * self.l0)
        self.Xtrain = Xtrain
        self.C = np.dot(Xtrain, Xtrain.T) * self.corrNorm
        self.CX = self.C.diagonal()
        self.D = torch.tensor(self.kernel(self.C.diagonal()[:,None], self.C, self.C.diagonal()[None,:]), requires_grad = False)
        if self.oneHot:
            self.Ytrain = tasks.oneHotEncoding(Ytrain)
        else:
            self.Ytrain = Ytrain
        tempYReshaped = self.Ytrain[:,0];
        for i in range(1,self.D):
            tempYReshaped = np.concatenate((tempYReshaped, self.Ytrain[:,i]))
        self.y = torch.tensor(tempYReshaped, dtype=torch.float32)

    def computeTestsetKernels(self, Xtest):
        self.Ptest = len(Xtest)
        self.C0 = np.dot(Xtest, Xtest.T).diagonal() * self.corrNorm
        self.C0X = np.dot(Xtest, self.Xtrain.T) * self.corrNorm 
        self.K0X = self.kernel(self.C0[:,None], self.C0X, self.CX[None, :])
        self.K0 = self.kernel(self.C0, self.C0, self.C0)
    
    def predict(self, Xtest):
        self.computeTestsetKernels(Xtest)
        rK = np.zeros((self.D*self.P, self.D* self.P)) #fast-varying index is mu.
        rK0 = np.zeros(self.D*self.Ptest)
        rK0X = np.zeros(( self.D*self.Ptest, self.D* self.P))
        for i in range(self.D):
            ind1 = i*self.P
            for j in range(self.D):
                ind2 = j*self.P
                rK[ind1:ind1+self.P,ind2:ind2+self.P] = self.optQ[i,j]*self.D/self.l1      
        for i in range(self.D):
            ind1 = i*self.Ptest
            rK0[ind1:ind1+self.Ptest] = self.optQ[i,i]*self.K0 /self.l1 
            for j in range(self.D):
                ind2 = j*self.P
                rK0X[ind1:ind1+self.Ptest,ind2:ind2+self.P] = self.optQ[i,j]*self.K0X  /self.l1
        self.rK0X = rK0X
        self.rK0 = rK0
        A = rK + (self.T) * np.eye(self.P*self.D)
        self.invK = np.linalg.inv(A)
        rKY = np.matmul(self.invK, self.y)
        self.prediction = np.matmul(rK0X, rKY)
        yPred = self.prediction.reshape(self.Ptest,self.D)
        return yPred

    def averageLoss(self, Ytest):
        if self.oneHot:
            self.Ytest = tasks.oneHotEncoding(Ytest, verbose=False)
        else:
            self.Ytest = Ytest
        tempYTestReshaped = self.Ytest[:,0]
        for i in range(1,self.D):
            tempYTestReshaped = np.concatenate((tempYTestReshaped, self.Ytest[:,i]))
        self.Ytest = torch.tensor(tempYTestReshaped)
        bias = (self.Ytest-self.prediction)**2
        biasPerClass = bias.reshape(self.Ptest,self.D).mean(axis = 0) # array of dimension k. each entry is average (on testset) bias for that class
        extVar = np.matmul(self.rK0X,np.matmul(self.invK, self.rK0X.T))
        var = self.rK0 - np.diag(extVar)
        varPerClass = var.reshape(self.Ptest,self.D).mean(axis = 0) # array of dimension k. each entry is average (on testset) var for that class
        predLoss = biasPerClass.mean().item() + varPerClass.mean().item()
        return predLoss, biasPerClass, varPerClass


import cupy as cp

### VERSION OF KERNELS THAT WORK WITH CUPY

def kernel_erf_cuda(cxx,cxy,cyy):
    return (2 / cp.pi)*cp.arcsin((2 * cxy) / cp.sqrt((1 + 2 * cxx)*(1 + 2 * cyy)))

def kernel_relu_cuda(cxx, cxy, cyy):
    u = cxy / cp.sqrt(cxx * cyy)
    kappa = (1 / (2 * cp.pi)) * (u * (cp.pi - cp.arccos(u)) + cp.sqrt(1 - u**2))
    return cp.sqrt(cxx * cyy) * kappa

def kernel_relu_bias_cuda(cxx, cxy, cyy):
    u = cxy / cp.sqrt(cxx * cyy)
    kappa = (1 / (2 * cp.pi)) * (u * (cp.pi - cp.arccos(u)) + cp.sqrt(1 - u**2))
    return cp.sqrt(cxx * cyy) * kappa 

def mean_relu_cuda(c):
    return cp.sqrt(c/2*cp.pi)

def kernel_id_cuda(cxx, cxy, cyy):
    return cxy    




import torch, torchvision, torchvision.transforms as t 
import numpy as np
import torch.nn as nn
import torch.nn.functional as F

def filter_by_label(data_loader, labels, P, dataSeed):
    data, target = next(iter(data_loader))
    mask = torch.zeros_like(target, dtype = torch.bool)
    for label in labels:
        mask |=  target  ==  label
    filtered_data = data[mask]
    filtered_labels = target[mask]
    rng = np.random.RandomState(dataSeed)  
    rp = rng.permutation(len(filtered_labels))
    filtered_data = filtered_data[rp[:P]]
    filtered_labels = filtered_labels[rp[:P]]
    #zero_one_labels = [0 if x  ==  labels[0] else 1 for x in filtered_labels]
    return filtered_data, filtered_labels #  torch.tensor(zero_one_labels)

def getTransforms(self):
        T = t.Compose([
            t.Resize(size = self.side_size), 
            t.ToTensor(), 
            t.Grayscale(), 
            #t.Normalize((0.5),(0.24)),
        ])
        if self.flatten:
            T = t.Compose([
                T, 
                t.Lambda(lambda x: torch.flatten(x))
            ])
        return T

def normalizeDataset(data, testData):
    mean = data.mean()
    std = data.std()
    data = data - mean
    testData = testData - mean
    data = data / std
    testData = testData /std
    return data, testData

def oneHotEncoding(y, verbose=True):
    uniqueClasses = torch.unique(y);
    classToIndex = {cls.item(): idx for idx, cls in enumerate(uniqueClasses)};
    mappedTensor = torch.tensor([classToIndex[val.item()] for val in y]);
    numClasses = len(uniqueClasses);
    oneHotEncoded = F.one_hot(mappedTensor, num_classes=numClasses);
    if verbose:
        print("Performing One Hot Encoding of your data:")
        print(f"> num of classes: {numClasses}")
        print(f"> class dictionary (from old to new):")
        for newClass, oldClass in enumerate(classToIndex):
            tempArray = np.zeros(numClasses);
            tempArray[newClass] = 1;
            print(f"   old class : {oldClass} >> new class {tempArray}")
    return oneHotEncoded;

class mnist_dataset: 
    def __init__(self, N, selectedLabels, dataSeed = 1234): #selectedLabels is a list, for example [2, 3] will select the labels 2 and 3 from mnist
        self.N = N
        self.side_size = int(np.sqrt(self.N))
        self.selectedLabels = selectedLabels
        self.dataSeed = dataSeed
    def make_data(self, P, Ptest, batchSize = 60000, flatten = False):
        self.flatten = flatten
        transformDataset = getTransforms(self)
        trainset = torchvision.datasets.MNIST(root = './data', train = True, download = True, transform = transformDataset)
        trainloader = torch.utils.data.DataLoader(trainset, batch_size = batchSize)
        testset = torchvision.datasets.MNIST(root = './data', train = False, download = True, transform = transformDataset)
        testloader = torch.utils.data.DataLoader(testset, batch_size = 10000, num_workers = 0)
        # Filter train and test datasets
        data, labels = filter_by_label(trainloader, self.selectedLabels, P, self.dataSeed)
        testData, testLabels = filter_by_label(testloader, self.selectedLabels, Ptest, self.dataSeed)
        data, testData  = normalizeDataset(data, testData)
        return data, labels.unsqueeze(1), testData, testLabels.unsqueeze(1)
 

class cifar_dataset: 
    def __init__(self, N, selectedLabels, dataSeed=123):
        self.N = N
        self.side_size = int(np.sqrt(self.N))
        self.selectedLabels = selectedLabels
        self.dataSeed = dataSeed
    def make_data(self, P, Ptest, batchSize = 60000, flatten = False):
        self.flatten = flatten
        transformDataset = getTransforms(self)
        trainset = torchvision.datasets.CIFAR10(root = './data', train = True, download = True, transform = transformDataset)
        trainloader = torch.utils.data.DataLoader(trainset, batch_size = batchSize, num_workers = 0)
        testset = torchvision.datasets.CIFAR10(root = './data', train = False, download = True, transform = transformDataset)
        testloader = torch.utils.data.DataLoader(testset, batch_size = 10000, num_workers = 0)
        #all_data, targets = next(iter(trainloader))
        # Filter train and test datasets
        data, labels = filter_by_label(trainloader, self.selectedLabels, P, self.dataSeed)
        testData, testLabels = filter_by_label(testloader, self.selectedLabels, Ptest, self.dataSeed)
        data, testData  = normalizeDataset(data, testData)
        return data, labels.unsqueeze(1), testData, testLabels.unsqueeze(1)



from .theory_1HL_FC_single_output import FC_1HL, FC_1HL_corrected, FC_1HL_nonodd
from .theory_1HL_FC_single_output_cuda import FC_1HL_cuda
from .theory_1HL_FC_multiple_output import FC_1HL_multiclass
from .theory_deep_FC import FC_deep
from .theory_1HL_CONV import CONV_1HL, validation_CONV_1HL
#from .theory_1HL_extras import FC_1HL_spectral, FC_1HL_multitask


import numpy as np
from scipy.optimize import minimize, fsolve
import torch
from .. import kernels

def kmatrix(Cmm,Cmn,Cnn,kernel):
    P1,P2 = Cmn.shape
    K = np.zeros((P1,P2))
    for i in range(P1): 
        for j in range(P2):         
            K[i][j] = kernel(Cmm[i][i], Cmn[i][j], Cnn[j][j])
            #K[j][i] = K[i][j]
    return K

class FC_1HL_spectral():
    def __init__(self, theta, N1, T, l0 = 1., l1 = 1., act = "erf"):
        self.N1 = N1
        self.l0 = l0
        self.l1 = l1 
        self.T = T
        self.kernel = eval(f"kernels.kernel_{act}")
        #self.theta = theta #fraction of shut nodes
        self.theta = theta #fraction of open nodes 

    def effectiveAction(self,y): #y is Q in the calculations
        #x = self.theta*(1+y)/(self.theta+(1-self.theta)*np.sqrt(1+y))
        x = self.theta/((1+y)**(1.5)*(1-self.theta + self.theta/np.sqrt(1+y)))
        A = self.T * np.identity(self.P) + (x/self.l1) * self.diagK
        invA = np.linalg.inv(A)
        return (- y*x - 2 * np.log(1-self.theta + (self.theta)/np.sqrt(1+y))
            + (1/self.N1) * np.sum(np.log(self.T + x*self.eigvalK/self.l1))
            + (1/self.N1) * np.dot(self.yT, np.dot(invA, self.yT)) )

    def minimizeAction(self, x0): # x0 is initial condition
        optQ = minimize(self.effectiveAction, x0, bounds = ((1e-8,np.inf),) , tol=1e-20)
        self.optQ = (optQ.x).item()
        self.optbarQ = self.theta/((1+self.optQ)**(1.5)*(1-self.theta + self.theta/np.sqrt(1+self.optQ)))
        print(f"value of S at the saddle point {self.effectiveAction(self.optQ)}")
        print(f"optQ is {self.optQ}")
        print(f"optbarQ is {self.optbarQ}")

    def preprocess(self, Xtrain, Ytrain):
        self.P, self.N0 = Xtrain.shape
        self.corrNorm = 1/(self.N0*self.l0)
        self.C = np.dot(Xtrain, Xtrain.T) * self.corrNorm
        self.Xtrain = Xtrain
        self.Ytrain = Ytrain
        self.CX = self.C.diagonal()
        self.K = kernels.computeKmatrix(self.C, self.kernel)
        [self.eigvalK, eigvecK] = np.linalg.eig(self.K)
        self.diagK = np.diagflat(self.eigvalK)
        self.Udag = eigvecK.T
        self.yT = np.matmul(self.Udag, Ytrain.squeeze())

    def computeTestsetKernels(self, Xtest):
        self.Ptest = len(Xtest)
        self.C0 = np.dot(Xtest, Xtest.T).diagonal() * self.corrNorm
        self.C0X = np.dot(Xtest, self.Xtrain.T) * self.corrNorm
        self.K0 =  self.kernel(self.C0, self.C0, self.C0) 
        self.K0X = self.kernel(self.C0[:,None], self.C0X, self.CX[None, :])
    
    def computeAveragePrediction(self, Xtest, Ytest):
        self.computeTestsetKernels(Xtest)
        self.orderParam = self.optQ / self.l1
        A = self.orderParam * self.K + (self.T) * np.eye(self.P)
        invK = np.linalg.inv(A)
        rK0 = self.orderParam * self.K0 
        rK0X = self.orderParam * self.K0X
        K0_invK = np.matmul(rK0X, invK)
        bias = Ytest - np.dot(K0_invK, self.Ytrain) 
        var = rK0 - np.sum(K0_invK * rK0X, axis=1)
        predLoss = bias**2 + var
        return predLoss.mean().item(), (bias**2).mean().item()
    
    def computePosteriorEigenvalues(self):
        norm = self.theta/np.sqrt(1+self.optQ) +1-self.theta
        self.p0 = (1-self.theta)/norm
        self.p1 = (self.theta/(np.sqrt(1+self.optQ)))/norm
        print(f"posterior prob of open node: {self.p1}")
        print(f"posterior prob of shut node: {self.p0}")


class FC_1HL_multitask():
    def __init__(self, N1, T, l0 = 1., l1 = 1., act = "erf"):
        self.N1 = N1
        self.l0 = l0
        self.l1 = l1 
        self.T = T
        self.kernel = eval(f"kernels.kernel_{act}")

    def effectiveAction(self, barQ, K1, K2, Kmix,labels):
        Ptot = self.Ps + self.Pt
        rK = torch.zeros((Ptot, Ptot), dtype=torch.float64)
        #assert (Kmix == KmixT).all()
        rK[:self.Ps, :self.Ps] = K1 * barQ[0]
        rK[self.Ps:Ptot, self.Ps:Ptot] = K2 * barQ[3]
        assert rK[self.Ps:Ptot, self.Ps:Ptot].shape == (self.Pt, self.Pt)
        rK[self.Ps:Ptot, :self.Ps] = Kmix.T * barQ[2]
        rK[:self.Ps, self.Ps:Ptot] = Kmix * barQ[1]
        A = rK / self.l1  +  self.T * torch.eye(Ptot) 
        invA = torch.inverse(A)
        return (  barQ[0] + barQ[3]
                - torch.log(barQ[0]*barQ[3]- barQ[1]*barQ[2])
                + (1/self.N1) * torch.matmul(labels, torch.matmul(invA, labels)) 
                + (1/self.N1) * torch.logdet(A)
                 )

    def computeGrad(self, barQ, K1, K2, Kmix, labels):
        barQ = torch.tensor(barQ, dtype = torch.float64, requires_grad = True)
        f = self.effectiveAction(barQ, K1, K2, Kmix, labels)
        f.backward(retain_graph=True)
        return barQ.grad.data.detach().numpy()   

    def computeAveragePrediction(self, inputs1, inputs2, labels1, labels2, testInputs1, testInputs2, testLabels1, testLabels2,kernel, args,theoryFile, barQ):    
        self.Ps = len(inputs1)
        self.Pt = len(inputs2)
        N0 = len(inputs1[0])  
        self.corrNorm = 1/(N0)
        #diag = self.gamma + self.l0
        #offDiag = -self.gamma
        offDiag = -self.gamma/2
        diag = self.l0
        regMat = [[diag, offDiag],[offDiag, diag]]
        invReg = np.linalg.inv(regMat)

        data = torch.concatenate((inputs1,inputs2))
        labels = torch.concatenate((labels1.squeeze().type(torch.DoubleTensor), labels2.squeeze().type(torch.DoubleTensor)))
        labels.requires_grad = False

        C1 = np.dot(inputs1,inputs1.T) * self.corrNorm * invReg[0][0]
        C2 = np.dot(inputs2,inputs2.T) * self.corrNorm * invReg[1][1]
        Cmix = np.dot(inputs1,inputs2.T) * self.corrNorm * invReg[0][1]

        K1 = torch.tensor(kmatrix(C1,C1,C1,kernel), dtype = torch.float64, requires_grad = False)
        K2 = torch.tensor(kmatrix(C2,C2,C2,kernel), dtype = torch.float64, requires_grad = False)
        Kmix = torch.tensor(kmatrix(C1,Cmix,C2,kernel), dtype = torch.float64, requires_grad = False)

        additionalArgs = (K1, K2, Kmix, args, labels)
        optQ = fsolve(self.computeGrad, barQ, args=additionalArgs, xtol = 1e-8)
        print(f"\nPoint where the gradient is approximately zero: {optQ}")
        isClose = np.isclose(self.scomputeGrad(optQ, *additionalArgs), [0.0, 0.0, 0.0, 0.0]) 
        print("\nis exact solution close to zero?", isClose)   
        #print(f"\nInteraction strength gamma: {self.gamma}")
        optQ = optQ.reshape(2,2)

        predLoss1,predLoss2 = 0,0
        for p in range(self.Ptest):
            x,y = np.array(testInputs1[p]), np.array(testLabels1[p])
            predLoss1 += self.computeSinglePrediction(data,labels,x,y,optQ, 1, kernel, K1, K2, Kmix, self.corrNorm, invReg)
            x,y = np.array(testInputs2[p]), np.array(testLabels2[p])
            predLoss2 += self.computeSinglePrediction(data,labels,x,y,optQ, 2, kernel, K1, K2, Kmix, self.corrNorm, invReg)

        with open(theoryFile, "a+") as f: 
            print(self.N1, self.Ps, self.Pt, self.T, self.gamma, self.l0, self.l1, optQ[0][0], optQ[0][1], optQ[1][1], predLoss1/self.Ptest, predLoss2/self.Ptest, self.Ptest, self.q, self.eta, self.rho, isClose.all(), file = f)

        print("test loss first task ", predLoss1 / self.Ptest , "\ntest loss second task ", predLoss2 / self.Ptest )

        return optQ
    
    def computeSinglePrediction(self, data, labels, x, y, optQ, task, kernel, K1, K2, Kmix, corrNorm, invReg):
        Ptot = self.Ps + self.Pt
        P = self.Ps
        corrXX = np.dot(x,x)*corrNorm*invReg[task-1][task-1]
        rKmu =  np.random.randn(Ptot)
        for mu in range(self.Ps):
            corrXData = np.dot(x,data[mu]) * corrNorm * invReg[task-1][0] 
            corrDataData = np.dot(data[mu],data[mu]) * corrNorm * invReg[0][0]
            rKmu[mu] = (optQ[task-1,0]/self.l1) * kernel(corrXX,corrXData,corrDataData) 
        for mu in range(self.Pt):
            corrXData2 = np.dot(x,data[self.Ps + mu]) * corrNorm * invReg[task-1][1]
            corrData2Data2 = np.dot(data[self.Ps+mu],data[self.Ps +mu]) * corrNorm * invReg[1][1]
            rKmu[self.Ps+mu] = (optQ[task-1,1]/self.l1) * kernel(corrXX,corrXData2,corrData2Data2)

        rKXX = (optQ[task-1,task-1]/self.l1) * kernel(corrXX,corrXX,corrXX) 
        rK = np.zeros((Ptot, Ptot), dtype=float)
        rK[:self.Ps, :self.Ps] = K1 * optQ[0, 0]
        rK[self.Ps:Ptot, self.Ps:Ptot] = K2 * optQ[1, 1]
        rK[self.Ps:Ptot, :self.Ps] = Kmix.T * optQ[0, 1]
        rK[:P, P:Ptot] = rK[P:Ptot, :P].T
        #rK[:P, P:Ptot] = rK[P:Ptot, :P]
        #assert rK[self.Ps:Ptot, self.Ps:Ptot].shape == (self.Pt, self.Pt)

        A = rK/self.l1 +  self.T * np.eye(Ptot) 
        invA = np.linalg.inv(A)
        rKmu_invA = np.matmul(rKmu, invA)
        bias = y - np.dot(rKmu_invA, labels)
        var = rKXX - np.dot(rKmu_invA, rKmu) 
        predErr = bias**2 + var
        return predErr.item()




from .kernels import kernel_erf
from .kernels import kernel_relu
from .kernels import kernel_relu_bias
from .kernels import mean_relu
from .kernels import computeKmatrix
from .kernels_torch import divide2dImage
from .kernels_torch import kernel_erf_torch
from .kernels_torch import kernel_relu_torch
from .kernels_torch import computeKmatrixTorch
from .kernels_torch import computeKmatrixMultipleCTorch
from .kernels_cuda import kernel_erf_cuda
from .kernels_cuda import kernel_relu_cuda
from .kernels_cuda import kernel_relu_bias_cuda
from .kernels_cuda import mean_relu_cuda




import numpy as np
import torch
from scipy.integrate import dblquad, quad

### VERSION OF KERNELS THAT WORK WITH TORCH. NEEDED WHEN GRADIENT OF THE ACTION CANNOT BE COMPUTED EXACTLY
def computeKmatrixTorch(C, kernel):
    diagonal = torch.diag(C).clone() #.detach().requires_grad_(True)
    diagonal_column = diagonal[:, None]  # Shape (n, 1) where n is the length of the diagonal
    diagonal_row = diagonal[None, :]     # Shape (1, n)
    k = kernel(diagonal_column, C.clone(), diagonal_row )#.detach().requires_grad_(True), diagonal_row)
    return k

def computeKmatrixMultipleCTorch(C1,C12,C2,  kernel):
    diagonal1 = torch.diag(C1).clone() #.detach().requires_grad_(True)
    diagonal2 = torch.diag(C2).clone()
    diagonal_column = diagonal1[:, None]  # Shape (n, 1) where n is the length of the diagonal
    diagonal_row = diagonal2[None, :]     # Shape (1, n)
    k = kernel(diagonal_column, C12.clone(), diagonal_row )#.detach().requires_grad_(True), diagonal_row)
    return k

def kernel_erf_torch(cxx,cxy,cyy):
    return (2 / torch.pi) * torch.arcsin((2 * cxy)/torch.sqrt((1 + 2 * cxx)*(1 + 2 * cyy)))

def kernel_relu_torch(cxx, cxy, cyy):
    u = cxy / torch.sqrt(cxx * cyy)
    kappa = (1 / (2 * torch.pi)) * (u * (torch.pi - torch.arccos(u)) + torch.sqrt(1 - u**2))
    return torch.sqrt(cxx * cyy) * kappa

def divide2dImage(array, k):
    n = array.shape[1]
    chunks = []
    for i in range(0, n, k):
        for j in range(0, n, k):
            chunk = array[i:i+k, j:j+k]
            chunks.append(chunk)
    return chunks


import numpy as np
import torch
from scipy.integrate import dblquad, quad
from functools import partial
import math

### VERSION OF KERNELS THAT WORK WITH NUMPY
def computeKmatrix(C, kernel):
    diagonal = np.diag(C)
    K = kernel(diagonal[:, None], C, diagonal[None, :]) # diagonal[:, None] makes it a column vector, diagonal[None, :] makes it a row vector
    return K

def kernel_erf(cxx,cxy,cyy):
    return (2 / np.pi)*np.arcsin((2 * cxy) / np.sqrt((1 + 2 * cxx)*(1 + 2 * cyy)))

def kernel_tanh(cxx: float, cxy: float, cyy: float):
    limit = 200
    output_shape = np.broadcast(cxx, cxy, cyy).shape
    #print(output_shape)
    # Initialize arrays to store results and errors
    result = np.empty(output_shape)
    error = np.empty(output_shape)
    # Perform integration for each element of the output shape
    for i in np.ndindex(output_shape):
        if len(i) == 1:
            cxx_val = cxx[i]
            cxy_val = cxy[i]
            cyy_val = cyy[i]
        else:
            cxx_val = cxx[i[0],0]
            cxy_val = cxy[i]
            cyy_val = cyy[0,i[1]]
        #cyy_val = cyy[indices]
        tildeC = np.array([[cxx_val, cxy_val], [cxy_val, cyy_val]])
        if cxx_val == cxy_val:
            def integralTanhKernelSingle(t1):
                #vec_t = np.array([t1, t2])
                gaussian_part = math.exp(-0.5 * t1**2 /cxx_val)
                tanh_part = math.tanh(t1)**2
                return gaussian_part * tanh_part
            result[i], error[i] = quad(integralTanhKernelSingle, -limit, limit)
            print(result[i], error[i])
            norm = 1/(2 * np.pi * np.sqrt(cxx_val))
        else:
            invTildeC = np.linalg.inv(tildeC)
            detTildeC = np.linalg.det(tildeC)
            def integralTanhKernel(t1, t2):
                vec_t = np.array([t1, t2])
                gaussian_part = math.exp(-0.5 * np.dot(vec_t.T, np.dot(invTildeC, vec_t)))
                tanh_part = math.tanh(t1) * math.tanh(t2)
                return gaussian_part * tanh_part
            result[i], error[i] = dblquad(integralTanhKernel, -limit, limit, -limit, limit)
            norm = 1 /(2 * np.pi *np.sqrt( detTildeC))
    print(result, error)
    return result * norm

#def integralTanhKernel(t1, t2, invTildeC):
#    vec_t = np.array([t1, t2])
#    gaussian_part = math.exp(-0.5 * np.dot(vec_t.T, np.dot(invTildeC, vec_t)))
#    tanh_part = math.tanh(t1) * math.tanh(t2)
#    return gaussian_part * tanh_part
#
#def integralTanhKernelSingle(t1, invTildeC):
#    #vec_t = np.array([t1, t2])
#    gaussian_part = math.exp(-0.5 * t1**2 * invTildeC)
#    tanh_part = math.tanh(t1)
#    return gaussian_part * tanh_part

def kernel_relu(cxx, cxy, cyy):
    u = cxy / np.sqrt(cxx * cyy)
    kappa = (1 / (2 * np.pi)) * (u * (np.pi - np.arccos(u)) + np.sqrt(1 - u**2))
    return np.sqrt(cxx * cyy) * kappa

def kernel_relu_bias(cxx, cxy, cyy):
    u = cxy / np.sqrt(cxx * cyy)
    kappa = (1 / (2 * np.pi)) * (u * (np.pi - np.arccos(u)) + np.sqrt(1 - u**2))
    return np.sqrt(cxx * cyy) * kappa 

def mean_relu(c):
    return np.sqrt(c/2*np.pi)

def kernel_id(cxx, cxy, cyy):
    return cxy    

#def kmatrix(C,kernel):
#    P = len(C)
#    K = np.zeros_like(C)
#    for i in range(P): 
#        for j in range(i,P):         
#            K[i][j] = kernel(C[i][i], C[i][j], C[j][j])
#            K[j][i] = K[i][j]
#    return K

def divide2dImage(array, k):
    n = array.shape[1]
    chunks = []
    for i in range(0, n, k):
        for j in range(0, n, k):
            chunk = array[i:i+k, j:j+k]
            chunks.append(chunk)
    return chunks


import numpy as np
from scipy.optimize import fsolve
from .. import kernels
import torch

class FC_deep():
    def __init__(self, 
                 L      : int, 
                 N1     : int, 
                 T      : float, 
                 priors : list = [1.0, 1.0], 
                 act    : str = "erf"):
        self.N1, self.T, self.L, self.l0, self.l1 = N1, T, L, priors[0],priors[1:] 
        self.kernelTorch = eval(f"kernels.kernel_{act}_torch")
        self.kernel = eval(f"kernels.kernel_{act}")
        
    def effectiveAction(self, Q):
        rKL = torch.tensor(self.C, dtype = torch.float64, requires_grad = False)
        for l in range(self.L):
            orderParam = Q[l] / self.l1[l]
            rKL = orderParam * self.kernelTorch(rKL.diagonal()[:,None], rKL, rKL.diagonal()[None,:])
        A = rKL +  self.T * torch.eye(self.P) 
        invA = torch.inverse(A)
        return ( torch.sum(Q - torch.log(Q))
                + (1/self.N1) * torch.matmul(self.y, torch.matmul(invA, self.y)) 
                + (1/self.N1) * torch.logdet(A)
                 )
    
    def computeActionGrad(self, Q):
        Q = torch.tensor(Q, dtype = torch.float64, requires_grad = True)
        f = self.effectiveAction(Q)
        f.backward()
        return Q.grad.data.detach().numpy()

    def optimize(self, Q0 = 1.):
        if isinstance(Q0, float):
            Q0 = np.ones(self.L)
        self.optQ = fsolve(self.computeActionGrad, Q0, xtol = 1e-8)
        isClose = np.isclose(self.computeActionGrad(self.optQ), np.zeros(self.L)) 
        self.converged = isClose.all()
        #print("\nis exact solution close to zero?", isClose)   
        #print(f"{self.L} hidden layer optQ is {self.optQ}")

    def setIW(self):
        self.optQ = np.ones(self.L)

    def preprocess(self, X, Y):
        self.X = X
        self.Y = Y
        self.P, self.N0 = X.shape
        self.corrNorm = 1/(self.N0 * self.l0)
        self.C = np.dot(X, X.T) * self.corrNorm
        self.CX = self.C.diagonal()
        self.y = Y.squeeze().to(torch.float64)
        self.y.requires_grad = False

    def computeTestsetKernels(self, X, Xtest):
        self.Ptest = len(Xtest)
        self.C0 = np.dot(Xtest, Xtest.T).diagonal() * self.corrNorm
        self.C0X = np.dot(Xtest, X.T) * self.corrNorm
    
    def predict(self, Xtest):
        self.computeTestsetKernels(self.X, Xtest)
        rKL = self.C
        rK0L = self.C0 
        rK0XL = self.C0X
        for l in range(self.L):
            orderParam = self.optQ[l] / self.l1[l]
            rKXL = rKL.diagonal() 
            rK0XL = orderParam * self.kernel(rK0L[:,None], rK0XL, rKXL[None, :])
            rK0L = orderParam * self.kernel(rK0L, rK0L, rK0L)
            rKL = orderParam * self.kernel(rKL.diagonal()[:,None], rKL, rKL.diagonal()[None,:])
        A = rKL + (self.T) * np.eye(self.P)
        invK = np.linalg.inv(A)
        K0_invK = np.matmul(rK0XL, invK)
        self.rK0L = rK0L
        self.K0_invK = K0_invK
        self.rK0XL = rK0XL
        self.Ypred = np.dot(K0_invK, self.Y)
        return self.Ypred
    
    def averageLoss(self, Ytest):
        bias = Ytest - self.Ypred  
        var = self.rK0L - np.sum(self.K0_invK * self.rK0XL, axis=1)
        predLoss = bias**2 + var
        return predLoss.mean().item(), (bias**2).mean().item(), var.mean().item()



from setuptools import setup, find_packages

VERSION = '0.1.11' 
DESCRIPTION = 'Bayesian deep neural networks in the proportional regime as renormalized kernel gaussian processes'
LONG_DESCRIPTION = 'Compute expected predictor of a Bayesian deep neural networks in the proportional regime in a non-parametric way using the equivalent GP'

setup(
        name='deepbays', 
        version=VERSION,
        author='Rosalba Pacelli',
        author_email='<rosalba.pacelli@gmail.com>',
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        packages=find_packages(),
        install_requires=['numpy', 'torch', 'torchvision', 'scipy'],
        
        keywords=['python', 'machine learning', 'gaussian processes', 'neural networks']
)